# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Similarity model used for metric learning.

Subclass of Keras.Model which provides methods for indexing, matching,
evaluation, and similarity training.

```python
from tf.keras import layers
from tensorflow_similarity.models import SimilarityModel

# Setup dataset using tf.sim samplers
train_ds = ...

# Create similarity loss
similarity_loss = ...

# Build the model
inputs = layers.Input(shape=(28,28,1))
x = layers.Conv2D(32, 3, activation='relu')(x)
x = layers.Flatten()(x)
outputs = layers.Dense(16, activation='relu')(x)

model = SimilarityModel(inputs, outputs)
model.compile(optimizer=Adam(LR), loss=similarity_loss)

history = model.fit(train_ds)
```
"""

from collections import defaultdict
from copy import copy
from pathlib import Path
from typing import DefaultDict, Dict, List, MutableMapping, MutableSequence
from typing import Optional, Sequence, Union
import numpy as np
from tabulate import tabulate
import tensorflow as tf
from tensorflow.keras.optimizers import Optimizer
from tensorflow.keras.metrics import Metric
from tensorflow.keras.losses import Loss
from tqdm.auto import tqdm

from tensorflow_similarity.classification_metrics import ClassificationMetric
from tensorflow_similarity.classification_metrics import make_classification_metric  # noqa
from tensorflow_similarity.distances import Distance
from tensorflow_similarity.distances import distance_canonicalizer
from tensorflow_similarity.training_metrics import DistanceMetric
from tensorflow_similarity.evaluators.evaluator import Evaluator
from tensorflow_similarity.indexer import Indexer
from tensorflow_similarity.losses import MetricLoss
from tensorflow_similarity.matchers import ClassificationMatch
from tensorflow_similarity.retrieval_metrics import RetrievalMetric
from tensorflow_similarity.stores import Store
from tensorflow_similarity.search import Search
from tensorflow_similarity.types import FloatTensor, Lookup, IntTensor, Tensor
from tensorflow_similarity.types import PandasDataFrame, CalibrationResults


@tf.keras.utils.register_keras_serializable(package="Similarity")
class SimilarityModel(tf.keras.Model):
    """Specialized Keras.Model which implement the core features needed for
    metric learning. In particular, `SimilarityModel()` supports indexing,
    searching and saving the embeddings predicted by the network.

    All Similarity models classes derive from this class to benefits from those
    core features.
    """

    # @property
    # def _index(self):
    #     if not hasattr(self, '_index'):
    #         ValueError("Index doesn't exist: index data before quering it")
    #     return self._index

    # @index.setter
    # def _index(self, index):
    #     self._index: Indexer = index

    def compile(
        self,
        optimizer: Union[Optimizer, str, Dict, List] = "rmsprop",
        loss: Optional[Union[Loss, MetricLoss, str, Dict, List]] = None,
        metrics: Optional[Union[Metric, DistanceMetric, str, Dict, List]] = None,  # noqa
        loss_weights: Optional[Union[List, Dict]] = None,
        weighted_metrics: Optional[Union[Metric, DistanceMetric, str, Dict, List]] = None,  # noqa
        run_eagerly: bool = False,
        steps_per_execution: int = 1,
        distance: Union[Distance, str] = "auto",
        embedding_output: Optional[int] = None,
        kv_store: Union[Store, str] = "memory",
        search: Union[Search, str] = "nmslib",
        evaluator: Union[Evaluator, str] = "memory",
        stat_buffer_size: int = 1000,
        **kwargs
    ):
        """Configures the model for training.

        Args:
            optimizer: String (name of optimizer) or optimizer instance. See
              [tf.keras.optimizers](https://www.tensorflow.org/api_docs/python/tf/keras/optimizers).

            loss: String (name of objective function), objective function, any
              `tensorflow_similarity.loss.*` instance or a `tf.keras.losses.Loss`
              instance. See the [Losses documentation](../losses.md) for a list of
              metric learning specific losses offered by TensorFlow Similairy and
              [tf.keras.losses](https://www.tensorflow.org/api_docs/python/tf/keras/losses)
              for the losses available directly in TensorFlow.

              metrics: List of metrics to be evaluated by the model during
              training and testing. Each of those can be a string, a function or a
              [tensorflow_similairty.metrics.*](../metrics.md) instance. Note that
              the metrics used for some type of metric-learning such as distance
              learning (e.g via triplet loss) have a different prototype than the
              metrics used in standard models and you can't use the
              `tf.keras.metrics` for those type of learning.

              Additionally many distance metrics are computed based of the
              [Indexer()](../indexer.md) performance. E.g Matching Top 1 accuracy.
              For technical and performance reasons, indexing data at each
              training batch to compute those is impractical so those metrics are
              computed at epoch end via the [EvalCallback](../callbacks.md)

              See [Evaluation Metrics](../eval_metrics.md) for a list of available
              metrics.

              For multi-output models you can specify different metrics for
              different outputs by passing a dictionary, such as
              `metrics={'similarity': 'min_neg_gap', 'other': ['accuracy',
              'mse']}`.  You can also pass a list (len = len(outputs)) of lists of
              metrics such as `metrics=[['min_neg_gap'], ['accuracy', 'mse']]` or
              `metrics=['min_neg_gap', ['accuracy', 'mse']]`. For outputs which
              are not related to metrics learning, you can use any of the standard
              `tf.keras.metrics`.

            loss_weights: Optional list or dictionary specifying scalar
              coefficients (Python floats) to weight the loss contributions of
              different model outputs. The loss value that will be minimized by
              the model will then be the *weighted sum* of all individual losses,
              weighted by the `loss_weights` coefficients.  If a list, it is
              expected to have a 1:1 mapping to the model's outputs. If a dict, it
              is expected to map output names (strings) to scalar coefficients.

            weighted_metrics: List of metrics to be evaluated and weighted by
              sample_weight or class_weight during training and testing.

            run_eagerly: Bool. Defaults to `False`. If `True`, this `Model`'s
              logic will not be wrapped in a `tf.function`. Recommended to leave
              this as `None` unless your `Model` cannot be run inside a
              `tf.function`.

            steps_per_execution: Int. Defaults to 1. The number of batches to
              run during each `tf.function` call. Running multiple batches inside
              a single `tf.function` call can greatly improve performance on TPUs
              or small models with a large Python overhead.  At most, one full
              epoch will be run each execution. If a number larger than the size
              of the epoch is passed,  the execution will be truncated to the size
              of the epoch.  Note that if `steps_per_execution` is set to `N`,
              `Callback.on_batch_begin` and `Callback.on_batch_end` methods will
              only be called every `N` batches (i.e. before/after each
              `tf.function` execution).

            distance: Distance used to compute embeddings proximity.  Defaults
              to 'cosine'.

            kv_store: How to store the indexed records.  Defaults to 'memory'.

            search: Which `Search()` framework to use to perform KNN search.
              Defaults to 'nmslib'.

            evaluator: What type of `Evaluator()` to use to evaluate index
              performance. Defaults to in-memory one.

            embedding_output: Which model output head predicts the embeddings
              that should be indexed. Default to None which is for single output
              model. For multi-head model, the callee, usually the
              `SimilarityModel()` class is responsible for passing the correct
              one.

            stat_buffer_size: Size of the sliding windows buffer used to compute
              index performance. Defaults to 1000.

        Raises:
            ValueError: In case of invalid arguments for
                `optimizer`, `loss` or `metrics`.
        """
        # Fetching the distance used from the first loss if auto
        if distance == "auto":
            if isinstance(loss, list):
                metric_loss = loss[0]
            else:
                metric_loss = loss

            try:
                distance = metric_loss.distance
            except AttributeError:
                msg = ("distance='auto' only works if the first loss is a "
                       "metric loss")

                raise ValueError(msg)
            print(
                f"Distance metric automatically set to {distance} use the "
                "distance arg to override.",
            )
        else:
            distance = distance_canonicalizer(distance)

        # init index
        self.create_index(
            distance=distance,
            search=search,
            kv_store=kv_store,
            evaluator=evaluator,
            embedding_output=embedding_output,
            stat_buffer_size=stat_buffer_size,
        )

        # call underlying keras method
        super().compile(
            optimizer=optimizer,
            loss=loss,
            metrics=metrics,
            loss_weights=loss_weights,
            weighted_metrics=weighted_metrics,
            run_eagerly=run_eagerly,
            steps_per_execution=steps_per_execution,
            **kwargs
        )

    def create_index(
        self,
        distance: Union[Distance, str] = "cosine",
        search: Union[Search, str] = "nmslib",
        kv_store: Union[Store, str] = "memory",
        evaluator: Union[Evaluator, str] = "memory",
        embedding_output: int = None,
        stat_buffer_size: int = 1000,
    ) -> None:
        """Create the model index to make embeddings searchable via KNN.

        This method is normally called as part of `SimilarityModel.compile()`.
        However, this method is provided if users want to define a custom index
        outside of the `compile()` method.

        NOTE: This method sets `SimilarityModel._index` and will replace any
        existing index.

        Args:
            distance: Distance used to compute embeddings proximity. Defaults to
            'auto'.

            kv_store: How to store the indexed records.  Defaults to 'memory'.

            search: Which `Search()` framework to use to perform KNN search.
            Defaults to 'nmslib'.

            evaluator: What type of `Evaluator()` to use to evaluate index
            performance. Defaults to in-memory one.

            embedding_output: Which model output head predicts the embeddings
            that should be indexed. Default to None which is for single output
            model. For multi-head model, the callee, usually the
            `SimilarityModel()` class is responsible for passing the correct
            one.

            stat_buffer_size: Size of the sliding windows buffer used to compute
            index performance. Defaults to 1000.

        Raises:
            ValueError: Invalid search framework or key value store.
        """
        # check if we we need to set the embedding head
        num_outputs = len(self.output_names)
        if embedding_output is not None and embedding_output > num_outputs:
            raise ValueError(
                "Embedding_output value exceed number of model outputs"
            )

        if embedding_output is None and num_outputs > 1:
            print(
                "Embedding output set to be model output 0. ",
                "Use the embedding_output arg to override this.",
            )
            embedding_output = 0

        # fetch embedding size as some ANN libs requires it for init
        if num_outputs > 1:
            self.embedding_size = self.outputs[embedding_output].shape[1]
        else:
            self.embedding_size = self.outputs[0].shape[1]

        self._index = Indexer(
            embedding_size=self.embedding_size,
            distance=distance,
            search=search,
            kv_store=kv_store,
            evaluator=evaluator,
            embedding_output=embedding_output,
            stat_buffer_size=stat_buffer_size,
        )

    def index(
        self,
        x: Tensor,
        y: IntTensor = None,
        data: Optional[Tensor] = None,
        build: bool = True,
        verbose: int = 1,
    ):
        """Index data.

        Args:
            x: Samples to index.

            y: class ids associated with the data if any. Defaults to None.

            store_data: store the data associated with the samples in the key
            value store. Defaults to True.

            build: Rebuild the index after indexing. This is needed to make the
            new samples searchable. Set it to false to save processing time
            when calling indexing repeatidly without the need to search between
            the indexing requests. Defaults to True.

            verbose: Output indexing progress info. Defaults to 1.
        """

        if not self._index:
            raise Exception(
                "You need to compile the model with a valid"
                "distance to be able to use the indexing"
            )
        if verbose:
            print("[Indexing %d points]" % len(x))
            print("|-Computing embeddings")
        predictions = self.predict(x)

        self._index.batch_add(
            predictions=predictions,
            labels=y,
            data=data,
            build=build,
            verbose=verbose,
        )

    def index_single(self,
                     x: Tensor,
                     y: IntTensor = None,
                     data: Optional[Tensor] = None,
                     build: bool = True,
                     verbose: int = 1):
        """Index data.

        Args:
            x: Sample to index.

            y: class id associated with the data if any. Defaults to None.

            data: store the data associated with the samples in the key
            value store. Defaults to None.

            build: Rebuild the index after indexing. This is needed to make the
            new samples searchable. Set it to false to save processing time
            when calling indexing repeatidly without the need to search between
            the indexing requests. Defaults to True.

            verbose: Output indexing progress info. Defaults to 1.
        """

        if not self._index:
            raise Exception('You need to compile the model with a valid'
                            'distance to be able to use the indexing')
        if verbose:
            print('[Indexing 1 point]')
            print('|-Computing embeddings')

        x = tf.expand_dims(x, axis=0)
        prediction = self.predict(x)
        self._index.add(prediction=prediction,
                        label=y,
                        data=data,
                        build=build,
                        verbose=verbose)

    def lookup(
        self, x: Tensor, k: int = 5, verbose: int = 1
    ) -> List[List[Lookup]]:
        """Find the k closest matches in the index for a set of samples.

        Args:
            x: Samples to match.

            k: Number of nearest neighboors to lookup. Defaults to 5.

            verbose: display progress. Default to 1.

        Returns
            list of list of k nearest neighboors:
            List[List[Lookup]]
        """
        predictions = self.predict(x)
        return self._index.batch_lookup(
            predictions=predictions, k=k, verbose=verbose
        )

    def single_lookup(self, x: Tensor, k: int = 5) -> List[Lookup]:
        """Find the k closest matches in the index for a given sample.

        Args:
            x: Sample to match.

            k: Number of nearest neighboors to lookup. Defaults to 5.

        Returns
            list of the k nearest neigboors info:
            List[Lookup]
        """
        x = tf.expand_dims(x, axis=0)
        prediction = self.predict(x)
        return self._index.single_lookup(prediction=prediction, k=k)

    def index_summary(self):
        "Display index info summary."
        self._index.print_stats()

    def calibrate(
        self,
        x: FloatTensor,
        y: IntTensor,
        thresholds_targets: MutableMapping[str, float] = {},
        k: int = 1,
        calibration_metric: Union[str, ClassificationMetric] = "f1",
        matcher: Union[str, ClassificationMatch] = "match_nearest",
        extra_metrics: MutableSequence[Union[str, ClassificationMetric]] = [
            "precision", "recall"
        ],  # noqa
        rounding: int = 2,
        verbose: int = 1,
    ) -> CalibrationResults:
        """Calibrate model thresholds using a test dataset.

        TODO: more detailed explaination.

        Args:

            x: examples to use for the calibration.

            y: labels associated with the calibration examples.

            thresholds_targets: Dict of performance targets to (if possible)
            meet with respect to the `calibration_metric`.

            calibration_metric:
            [ClassificationMetric()](classification_metrics/overview.md) used
            to evaluate the performance of the index.

            k: How many neighboors to use during the calibration.
            Defaults to 1.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold. Defaults to
            'match_nearest'.

            extra_metrics: List of additional
            `tf.similarity.classification_metrics.ClassificationMetric()`
            to compute and report. Defaults to ['precision', 'recall'].


            rounding: Metric rounding. Default to 2 digits.

            verbose: Be verbose and display calibration results.
            Defaults to 1.

        Returns:
            CalibrationResults containing the thresholds and cutpoints Dicts.
        """

        # predict
        predictions = self.predict(x)

        # calibrate
        return self._index.calibrate(
            predictions=predictions,
            target_labels=y,
            thresholds_targets=thresholds_targets,
            k=k,
            calibration_metric=calibration_metric,
            matcher=matcher,
            extra_metrics=extra_metrics,
            rounding=rounding,
            verbose=verbose,
        )

    def match(
        self,
        x: FloatTensor,
        cutpoint="optimal",
        no_match_label=-1,
        k=1,
        matcher: Union[str, ClassificationMatch] = "match_nearest",
        verbose=0,
    ):
        """Match a set of examples against the calibrated index

        For the match function to work, the index must be calibrated using
        calibrate().

        Args:
            x: Batch of examples to be matched against the index.

            cutpoint: Which calibration threshold to use.
            Defaults to 'optimal' which is the optimal F1 threshold computed
            using calibrate().

            no_match_label: Which label value to assign when there is no
            match. Defaults to -1.

            k: How many neighboors to use during the calibration.
            Defaults to 1.

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.

            verbose. Be verbose. Defaults to 0.

        Returns:
            List of class ids that matches for each supplied example

        Notes:
            This function matches all the cutpoints at once internally as there
            is little performance downside to do so and allows to do the
            evaluation in a single go.

        """
        # basic checks
        if not self._index.is_calibrated:
            raise ValueError("Uncalibrated model: run model.calibration()")

        # get predictions
        predictions = self.predict(x)

        # matching
        matches = self._index.match(
            predictions,
            no_match_label=no_match_label,
            k=k,
            matcher=matcher,
            verbose=verbose,
        )

        # select which matches to return
        if cutpoint == "all":  # returns all the cutpoints for eval purpose.
            return matches
        else:  # normal match behavior - returns a specific cut point
            return matches[cutpoint]

    def evaluate_retrieval(
        self,
        x: Tensor,
        y: IntTensor,
        retrieval_metrics: Sequence[RetrievalMetric],  # noqa
        verbose: int = 1,
    ) -> Dict[str, np.ndarray]:
        """Evaluate the quality of the index against a test dataset.

        Args:
            x: Examples to be matched against the index.

            y: Label associated with the examples supplied.

            retrieval_metrics: List of
            [RetrievalMetric()](retrieval_metrics/overview.md) to compute.

            verbose (int, optional): Display results if set to 1 otherwise
            results are returned silently. Defaults to 1.

        Returns:
            Dictionary of metric results where keys are the metric names and
            values are the metrics values.

        Raises:
            IndexError: Index must contain embeddings but is currently empty.
        """
        if self._index.size() == 0:
            raise IndexError("Index must contain embeddings but is "
                             "currently empty. Have you run model.index()?")

        # get embeddings
        if verbose:
            print("|-Computing embeddings")
        predictions = self.predict(x)

        if verbose:
            print("|-Computing retrieval metrics")

        results = self._index.evaluate_retrieval(
            predictions=predictions,
            target_labels=y,
            retrieval_metrics=retrieval_metrics,
            verbose=verbose,
        )

        if verbose:
            table = zip(results.keys(), results.values())
            headers = ["metric", "Value"]
            print("\n [Summary]\n")
            print(tabulate(table, headers=headers))

        return results

    def evaluate_classification(
        self,
        x: Tensor,
        y: IntTensor,
        k: int = 1,
        extra_metrics: MutableSequence[Union[str, ClassificationMetric]] = [
            "precision", "recall"
        ],  # noqa
        matcher: Union[str, ClassificationMatch] = "match_nearest",
        verbose: int = 1,
    ) -> DefaultDict[str, Dict[str, Union[str, np.ndarray]]]:
        """Evaluate model classification matching on a given evaluation dataset.

        Args:
            x: Examples to be matched against the index.

            y: Label associated with the examples supplied.

            k: How many neighbors to use to perform the evaluation.
            Defaults to 1.

            extra_metrics: List of additional
            `tf.similarity.classification_metrics.ClassificationMetric()` to
            compute and report. Defaults to ['precision', 'recall'].

            matcher: {'match_nearest', 'match_majority_vote'} or
            ClassificationMatch object. Defines the classification matching,
            e.g., match_nearest will count a True Positive if the query_label
            is equal to the label of the nearest neighbor and the distance is
            less than or equal to the distance threshold.

            verbose (int, optional): Display results if set to 1 otherwise
            results are returned silently. Defaults to 1.

        Returns:
            Dictionary of (distance_metrics.md)[evaluation metrics]

        Raises:
            IndexError: Index must contain embeddings but is currently empty.
            ValueError: Uncalibrated model: run model.calibration()")
        """
        # There is some code duplication in this function but that is the best
        # solution to keep the end-user API clean and doing inferences once.
        if self._index.size() == 0:
            raise IndexError("Index must contain embeddings but is "
                             "currently empty. Have you run model.index()?")

        if not self._index.is_calibrated:
            raise ValueError("Uncalibrated model: run model.calibration()")

        cal_metric = self._index.get_calibration_metric()

        # get embeddings
        if verbose:
            print("|-Computing embeddings")
        predictions = self.predict(x)

        results: DefaultDict[
            str, Dict[str, Union[str, np.ndarray]]
        ] = defaultdict(dict)

        if verbose:
            pb = tqdm(
                total=len(self._index.cutpoints), desc="Evaluating cutpoints"
            )

        for cp_name, cp_data in self._index.cutpoints.items():
            # create a metric that match at the requested k and threshold
            distance_threshold = float(cp_data["distance"])
            metric = make_classification_metric(cal_metric.name)
            metrics = copy(extra_metrics)
            metrics.append(metric)

            res: Dict[str, Union[str, np.ndarray]] = {}
            res.update(
                self._index.evaluate_classification(
                    predictions,
                    y,
                    [distance_threshold],
                    metrics=metrics,
                    matcher=matcher,
                    k=k,
                )
            )
            res["distance"] = tf.constant([distance_threshold])
            res["name"] = cp_name
            results[cp_name] = res
            if verbose:
                pb.update()

        if verbose:
            pb.close()

        if verbose:
            headers = ["name", cal_metric.name]
            for i in results["optimal"].keys():
                if i not in headers:
                    headers.append(str(i))
            rows = []
            for data in results.values():
                rows.append([data[v] for v in headers])
            print("\n [Summary]\n")
            print(tabulate(rows, headers=headers))

        return results

    def reset_index(self):
        "Reinitialize the index"
        self._index.reset()

    def index_size(self) -> int:
        "Return the index size"
        return self._index.size()

    def load_index(self, filepath: str):
        """Load Index data from a checkpoint and initialize underlying
        structure with the reloaded data.

        Args:
            path: Directory where the checkpoint is located.
            verbose: Be verbose. Defaults to 1.
        """

        index_path = Path(filepath) / "index"
        self._index = Indexer.load(index_path)

    def save_index(self, filepath, compression=True):
        """Save the index to disk

        Args:
            path: directory where to save the index
            compression: Store index data compressed. Defaults to True.
        """
        index_path = Path(filepath) / "index"
        self._index.save(index_path, compression=compression)

    def save(
        self,
        filepath: str,
        save_index: bool = True,
        compression: bool = True,
        overwrite: bool = True,
        include_optimizer: bool = True,
        signatures=None,
        options=None,
        save_traces: bool = True,
    ):
        """Save the model and the index.

        Args:
            filepath: where to save the model.

            save_index: Save the index content. Defaults to True.

            compression: Compress index data. Defaults to True.

            overwrite: Overwrite previous model. Defaults to True.

            include_optimizer: Save optimizer state. Defaults to True.

            signatures: Signatures to save with the model. Defaults to None.

            options: A `tf.saved_model.SaveOptions` to save with the model.
            Defaults to None.

            save_traces (optional): When enabled, the SavedModel will
            store the function traces for each layer. This can be disabled,
            so that only the configs of each layer are stored.
            Defaults to True. Disabling this will decrease serialization
            time and reduce file size, but it requires that all
            custom layers/models implement a get_config() method.
        """

        # save trace doesn't exist prior to 2.4 -- asking for it but not
        # using it

        # call underlying keras method to save the mode graph and its weights
        tf.keras.models.save_model(
            self,
            filepath,
            overwrite=overwrite,
            include_optimizer=include_optimizer,
            signatures=signatures,
            options=options,
            save_traces=save_traces,
        )
        if hasattr(self, "_index") and self._index and save_index:
            self.save_index(filepath, compression=compression)
        else:
            print("Index not saved as save_index=False")

    def to_data_frame(self, num_items: int = 0) -> PandasDataFrame:
        """Export data as pandas dataframe

        Args:
            num_items (int, optional): Num items to export to the dataframe.
            Defaults to 0 (unlimited).

        Returns:
            pd.DataFrame: a pandas dataframe.
        """
        return self._index.to_data_frame(num_items=num_items)

    # We don't need from_config as the index is reloaded separatly.
    # this is kept as a reminder that it was looked into and decided to split
    # the index reloading instead of overloading this method.
    # @classmethod
    # def from_config(cls, config):
    #     return super().from_config(**config)



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"EfficientNet backbone for similarity learning"
import re
from typing import Tuple, Callable, Union
import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras.applications import efficientnet
from tensorflow_similarity.layers import MetricEmbedding
from tensorflow_similarity.models import SimilarityModel

EFF_INPUT_SIZE = {
    "B0": 224,
    "B1": 240,
    "B2": 260,
    "B3": 300,
    "B4": 380,
    "B5": 456,
    "B6": 528,
    "B7": 600,
}

EFF_ARCHITECTURE = {
    "B0": efficientnet.EfficientNetB0,
    "B1": efficientnet.EfficientNetB1,
    "B2": efficientnet.EfficientNetB2,
    "B3": efficientnet.EfficientNetB3,
    "B4": efficientnet.EfficientNetB4,
    "B5": efficientnet.EfficientNetB5,
    "B6": efficientnet.EfficientNetB6,
    "B7": efficientnet.EfficientNetB7,
}


# Create an image augmentation pipeline.
def EfficientNetSim(input_shape: Tuple[int],
                    embedding_size: int = 128,
                    variant: str = "B0",
                    weights: str = "imagenet",
                    augmentation: Union[Callable, str] = "basic",
                    trainable: str = "frozen",
                    l2_norm: bool = True):
    """Build an EffecientNet Model backbone for similarity learning

    Architecture from [EfficientNet: Rethinking Model Scaling for Convolutional Neural Networks
](https://arxiv.org/abs/1905.11946)

    Args:
        input_shape: Size of the image input prior to augmentation,
        must be bigger than the size of Effnet version you use. See below for
        min input size.

        embedding_size: Size of the output embedding. Usually between 64
        and 512. Defaults to 128.
        variant: Which Variant of the EfficientNEt to use. Defaults to "B0".

        weights: Use pre-trained weights - the only available currently being
        imagenet. Defaults to "imagenet".

        augmentation: How to augment the data - either pass a Sequential model
        of keras.preprocessing.layers or use the built in one or set it to
        None to disable. Defaults to "basic".

        trainable: Make the EfficienNet backbone fully trainable or partially
        trainable. Either "full" to make the entire backbone trainable,
        "partial" to only make the last 3 block trainable or "frozen" to make
        it not trainable. Defaults to "frozen".

        l2_norm: If True, tensorflow_similarity.layers.MetricEmbedding is used
        as the last layer, otherwise keras.layers.Dense is used. This should be
        true when using cosine distance. Defaults to True.

    Note:
        EfficientNet expects images at the following size:
         - "B0": 224,
         - "B1": 240,
         - "B2": 260,
         - "B3": 300,
         - "B4": 380,
         - "B5": 456,
         - "B6": 528,
         - "B7": 600,

    """

    # input
    inputs = layers.Input(shape=input_shape)
    x = inputs

    if variant not in EFF_INPUT_SIZE:
        raise ValueError("Unknown efficientnet variant. Valid B0...B7")
    img_size = EFF_INPUT_SIZE[variant]

    # augmentation
    if augmentation == "basic":
        # augs usually used in benchmark and work almost always well
        augmentation_layers = tf.keras.Sequential([
            layers.experimental.preprocessing.RandomCrop(img_size, img_size),
            layers.experimental.preprocessing.RandomFlip("horizontal")
        ])
    else:
        augmentation_layers = augmentation

    # add the basic version or the suppplied one.
    if augmentation:
        x = augmentation_layers(x)

    x = build_effnet(x, variant, weights, trainable)
    x = layers.GlobalAveragePooling2D()(x)
    if l2_norm:
        outputs = MetricEmbedding(embedding_size)(x)
    else:
        outputs = layers.Dense(embedding_size)(x)
    return SimilarityModel(inputs, outputs)


def build_effnet(x, variant, weights, trainable):
    "Build the requested efficient net."

    # init
    effnet_fn = EFF_ARCHITECTURE[variant]
    effnet = effnet_fn(weights=weights, include_top=False)

    if trainable == "full":
        effnet.trainable = True
    elif trainable == "partial":
        # let's mark the top part of the network as trainable
        effnet.trainable = True
        for layer in effnet.layers:
            # Freeze all the layers before the the last 3 blocks
            if not re.search("^block[5,6,7]|^top", layer.name):
                layer.trainable = False
            # don't change the batchnorm weights
            if isinstance(layer, layers.BatchNormalization):
                layer.trainable = False
    elif trainable=='frozen':
        effnet.trainable = False
    else:
        raise ValueError(f"{trainable} is not a supported option for 'trainable'.")

    # wire
    x = efficientnet.preprocess_input(x)
    x = effnet(x)

    return x



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys

from setuptools import find_packages
from setuptools import setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


# We use the same setup.py for both tensorflow_similarity and tfsim-nightly
# packages. The package is controlled from the argument line when building the
# pip package.
project_name = "tensorflow_similarity"
if "--project_name" in sys.argv:
    project_name_idx = sys.argv.index("--project_name")
    project_name = sys.argv[project_name_idx + 1]
    sys.argv.remove("--project_name")
    sys.argv.pop(project_name_idx)


setup(
    name=project_name,
    version=get_version("tensorflow_similarity/__init__.py"),
    description="Metric Learning for Humans",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Tensorflow Similarity authors",
    author_email="tf-similarity@google.com",
    url="https://github.com/tensorflow/similarity",
    license="Apache License 2.0",
    install_requires=[
        "bokeh",
        "distinctipy",
        "matplotlib",
        "nmslib",
        "numpy",
        "pandas",
        "Pillow",
        "tabulate",
        "tensorflow>=2.4",
        "tensorflow-datasets>=4.2",
        "tqdm",
        "umap-learn",
    ],
    extras_require={
        "dev": [
            "flake8",
            "mkdocs",
            "mkdocs-autorefs",
            "mkdocs-material",
            "mkdocstrings",
            "mypy",
            "pytest",
            "pytype",
            "setuptools",
            "twine",
            "types-tabulate",
            "wheel",
        ],
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    packages=find_packages(),
)



import re
from typing import Tuple

import numpy as np
import pytest
import tensorflow as tf

from tensorflow_similarity.matchers import ClassificationMatch
from tensorflow_similarity.types import FloatTensor, IntTensor

attributes = ("tp", "fp", "tn", "fn", "count")


class ConcreteClassificationMatch(ClassificationMatch):
    def derive_match(
        self, lookup_labels: IntTensor, lookup_distances: FloatTensor
    ) -> Tuple[FloatTensor, FloatTensor]:
        return lookup_labels, lookup_distances


class BadClassificationMatch(ClassificationMatch):
    "Derive match should return 2D tensors, but here we return 1D."

    def derive_match(
        self, lookup_labels: IntTensor, lookup_distances: FloatTensor
    ) -> Tuple[FloatTensor, FloatTensor]:
        return (
            tf.reshape(lookup_labels, (-1,)),
            tf.reshape(lookup_distances, (-1,)),
        )


def test_compile():
    cm = ConcreteClassificationMatch(name="foo", canonical_name="bar")

    # Pass distance_thresholds as a 1D tensor.
    distance_thresholds = tf.constant([1, 2, 3, 4, 5, 6, 7])
    cm.compile(distance_thresholds=distance_thresholds)

    expected_dt = tf.cast(distance_thresholds, dtype="float32")
    assert cm.name == "foo"
    assert cm.canonical_name == "bar"
    assert tf.math.reduce_all(
        tf.shape(cm.distance_thresholds) == tf.shape(expected_dt)
    )
    assert tf.math.reduce_all(
        tf.math.equal(cm.distance_thresholds, expected_dt)
    )


def test_compute_match_indicators():
    cm = ConcreteClassificationMatch(name="foo", canonical_name="bar")

    # Pass distance_thresholds as a 1D tensor.
    distance_thresholds = tf.constant([1.0, 2.0])
    cm.compile(distance_thresholds=distance_thresholds)

    query_labels = tf.constant([10, 20, 10, 20])
    lookup_labels = tf.constant([[10], [20], [30], [40]])
    lookup_distances = tf.constant([[1.0], [1.0], [2.0], [2.0]])

    match_mask, distance_mask = cm._compute_match_indicators(
        query_labels, lookup_labels, lookup_distances
    )

    np.testing.assert_array_equal(
        match_mask.numpy(), np.array([[True], [True], [False], [False]])
    )

    np.testing.assert_array_equal(
        distance_mask.numpy(),
        np.array([[True, True], [True, True], [False, True], [False, True]]),
    )


def test_compute_match_indicators_1d():
    """Check that we handle 1D derive match results."""
    cm = BadClassificationMatch(name="foo", canonical_name="bar")

    # Pass distance_thresholds as a 1D tensor.
    distance_thresholds = tf.constant([1.0, 2.0])
    cm.compile(distance_thresholds=distance_thresholds)

    query_labels = tf.constant([10, 20, 10, 20])
    lookup_labels = tf.constant([[10], [20], [30], [40]])
    lookup_distances = tf.constant([[1.0], [1.0], [2.0], [2.0]])

    match_mask, distance_mask = cm._compute_match_indicators(
        query_labels, lookup_labels, lookup_distances
    )

    np.testing.assert_array_equal(
        match_mask.numpy(), np.array([[True], [True], [False], [False]])
    )

    np.testing.assert_array_equal(
        distance_mask.numpy(),
        np.array([[True, True], [True, True], [False, True], [False, True]]),
    )


def test_compute_count():
    cm = ConcreteClassificationMatch(name="foo", canonical_name="bar")

    # Pass distance_thresholds as a 1D tensor.
    distance_thresholds = tf.constant([1.0, 2.0])
    cm.compile(distance_thresholds=distance_thresholds)

    query_labels = tf.constant([10, 20, 10, 20])
    lookup_labels = tf.constant([[10], [20], [30], [40]])
    lookup_distances = tf.constant([[1.0], [1.0], [2.0], [2.0]])

    cm.compute_count(query_labels, lookup_labels, lookup_distances)

    np.testing.assert_array_equal(cm.tp.numpy(), np.array([2, 2]))
    np.testing.assert_array_equal(cm.fp.numpy(), np.array([0, 2]))
    np.testing.assert_array_equal(cm.tn.numpy(), np.array([2, 0]))
    np.testing.assert_array_equal(cm.fn.numpy(), np.array([0, 0]))
    assert cm.count == 4


@pytest.mark.parametrize("attribute", attributes, ids=attributes)
def test_attribute_asserts(attribute):
    """Uninitialized attrs should through a ValueError."""
    cm = ConcreteClassificationMatch(name="foo", canonical_name="bar")

    msg = "Matcher.compute_count() must be called before accessing the counts."

    with pytest.raises(AttributeError, match=re.escape(msg)):
        getattr(cm, attribute)


def test_check_shape_valid():
    cm = ConcreteClassificationMatch(name="foo", canonical_name="bar")

    queries = tf.constant([[1], [2], [3]])
    ll = tf.constant([[1], [2], [3]])
    ld = tf.constant([[0.1], [0.2], [0.3]])

    assert cm._check_shape(queries, ll, ld)


def test_check_shape_invalid_queries():
    cm = ConcreteClassificationMatch(name="foo", canonical_name="bar")

    queries = tf.constant([[1], [2], [3], [4]])
    ll = tf.constant([[1], [2], [3]])
    ld = tf.constant([[0.1], [0.2], [0.3]])

    msg = "Number of query labels must match the number of lookup_label sets."

    with pytest.raises(ValueError, match=re.escape(msg)):
        cm._check_shape(queries, ll, ld)


def test_check_shape_invalid_lookup_rank():
    cm = ConcreteClassificationMatch(name="foo", canonical_name="bar")

    queries = tf.constant([[1], [2], [3]])
    ll = tf.constant([1, 2, 3])
    ld = tf.constant([[0.1], [0.2], [0.3]])

    msg = "lookup_labels must be a 2D tensor of shape [len(query_labels), K]."

    with pytest.raises(ValueError, match=re.escape(msg)):
        cm._check_shape(queries, ll, ld)


def test_check_shape_invalid_distance_rank():
    cm = ConcreteClassificationMatch(name="foo", canonical_name="bar")

    queries = tf.constant([[1], [2], [3]])
    ll = tf.constant([[1], [2], [3]])
    ld = tf.constant([0.1, 0.2, 0.3])

    msg = (
        "lookup_distances must be a 2D tensor of shape "
        "[len(query_labels), K]."
    )

    with pytest.raises(ValueError, match=re.escape(msg)):
        cm._check_shape(queries, ll, ld)


def test_check_shape_labels_dist_mismatch():
    cm = ConcreteClassificationMatch(name="foo", canonical_name="bar")

    queries = tf.constant([[1], [2], [3]])
    ll = tf.constant([[1], [2], [3]])
    ld = tf.constant([[0.1, 0.2], [0.2, 0.3], [0.3, 0.4]])

    msg = (
        "Number of number of lookup labels must match the number "
        "of lookup distances."
    )

    with pytest.raises(ValueError, match=re.escape(msg)):
        cm._check_shape(queries, ll, ld)



import math

import numpy as np
import tensorflow as tf

from tensorflow_similarity.models import SimilarityModel
from tensorflow_similarity.callbacks import EvalCallback
from tensorflow_similarity.callbacks import SplitValidationLoss
from tensorflow_similarity.evaluators import MemoryEvaluator
from tensorflow_similarity.classification_metrics import Recall


def test_eval_init_defaults():
    queries = tf.constant([[1, 2], [3, 4]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1, 2], [3, 4]])
    target_labels = tf.constant([1, 2])

    callback = EvalCallback(
        queries=queries,
        query_labels=query_labels,
        targets=targets,
        target_labels=target_labels,
    )

    assert tf.math.reduce_all(callback.queries == queries)
    assert tf.math.reduce_all(callback.query_labels == query_labels)
    assert tf.math.reduce_all(callback.targets == targets)
    assert tf.math.reduce_all(callback.target_labels == target_labels)
    assert callback.distance == 'cosine'
    assert isinstance(callback.evaluator, MemoryEvaluator)
    assert {'binary_accuracy', 'f1score'} == set([m.name for m in callback.metrics])
    assert callback.k == 1
    assert tf.math.reduce_all(
        callback.distance_thresholds == tf.constant([math.inf]))
    assert callback.matcher == 'match_nearest'
    assert callback.tb_writer is None


def test_eval_init(tmp_path):
    queries = tf.constant([[1, 2], [3, 4]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1, 2], [3, 4]])
    target_labels = tf.constant([1, 2])
    distance = 'l2'
    metrics = [Recall()]
    k = 11
    matcher = 'majority_vote'
    distance_thresholds = tf.constant([0.1, 0.2, 0.3])

    log_dir = tmp_path / 'sec/'
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = EvalCallback(queries=queries,
                            query_labels=query_labels,
                            targets=targets,
                            target_labels=target_labels,
                            distance=distance,
                            metrics=metrics,
                            k=k,
                            matcher=matcher,
                            distance_thresholds=distance_thresholds,
                            tb_logdir=log_dir)

    assert tf.math.reduce_all(callback.queries == queries)
    assert tf.math.reduce_all(callback.query_labels == query_labels)
    assert tf.math.reduce_all(callback.targets == targets)
    assert tf.math.reduce_all(callback.target_labels == target_labels)
    assert callback.distance == 'l2'
    assert isinstance(callback.evaluator, MemoryEvaluator)
    assert {'recall'} == set([m.name for m in callback.metrics])
    assert callback.k == 11
    assert tf.math.reduce_all(
        callback.distance_thresholds == distance_thresholds)
    assert callback.matcher == 'majority_vote'
    assert isinstance(callback.tb_writer, tf.summary.SummaryWriter)


def test_eval_callback(tmp_path):
    queries = tf.constant([[1, 2], [1, 2]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1, 2], [1, 2]])
    target_labels = tf.constant([1, 2])

    log_dir = tmp_path / 'sec/'
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = EvalCallback(queries=queries,
                            query_labels=query_labels,
                            targets=targets,
                            target_labels=target_labels,
                            tb_logdir=str(log_dir))

    # manually set model ^^
    tf.keras.backend.clear_session()
    inputs = tf.keras.layers.Input(shape=(2,))
    outputs = tf.keras.layers.Dense(2)(inputs)
    model = SimilarityModel(inputs, outputs)
    model.compile('adam', loss='mse', distance='cosine')
    callback.model = model

    # call the only callback method implemented
    logs = {}
    callback.on_epoch_end(0, logs)

    metric_values = np.array(list(logs.values()))
    assert {'binary_accuracy', 'f1score'} == logs.keys()
    np.testing.assert_allclose(np.array([0.5, 0.666667]),
                               metric_values,
                               rtol=1e-5,
                               atol=0)


def test_split_val_init_defaults():
    queries = tf.constant([[1, 2], [3, 4]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1, 2], [3, 4]])
    target_labels = tf.constant([1, 2])
    known_classes = tf.constant([1])
    q_known = tf.constant([[1, 2]])
    q_label_known = tf.constant([1], dtype='int32')
    q_unknown = tf.constant([[3, 4]])
    q_label_unknown = tf.constant([2], dtype='int32')

    callback = SplitValidationLoss(
        queries=queries,
        query_labels=query_labels,
        targets=targets,
        target_labels=target_labels,
        known_classes=known_classes,
    )

    assert tf.math.reduce_all(callback.targets == targets)
    assert tf.math.reduce_all(callback.target_labels == target_labels)
    assert callback.distance == 'cosine'
    assert isinstance(callback.evaluator, MemoryEvaluator)
    assert {'binary_accuracy', 'f1score'} == set([m.name for m in callback.metrics])
    assert callback.k == 1
    assert tf.math.reduce_all(
        callback.distance_thresholds == tf.constant([math.inf]))
    assert callback.matcher == 'match_nearest'
    assert callback.tb_writer is None

    assert tf.math.reduce_all(callback.queries_known == q_known)
    assert tf.math.reduce_all(callback.query_labels_known == q_label_known)
    assert tf.math.reduce_all(callback.queries_unknown == q_unknown)
    assert tf.math.reduce_all(callback.query_labels_unknown == q_label_unknown)


def test_split_val_init(tmp_path):
    queries = tf.constant([[1, 2], [3, 4]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1, 2], [3, 4]])
    target_labels = tf.constant([1, 2])
    known_classes = tf.constant([1])
    distance = 'l2'
    metrics = [Recall()]
    k = 11
    matcher = 'majority_vote'
    distance_thresholds = tf.constant([0.1, 0.2, 0.3])
    q_known = tf.constant([[1, 2]])
    q_label_known = tf.constant([1], dtype='int32')
    q_unknown = tf.constant([[3, 4]])
    q_label_unknown = tf.constant([2], dtype='int32')

    log_dir = tmp_path / 'sec/'
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = SplitValidationLoss(queries=queries,
                                   query_labels=query_labels,
                                   targets=targets,
                                   target_labels=target_labels,
                                   known_classes=known_classes,
                                   distance=distance,
                                   metrics=metrics,
                                   k=k,
                                   matcher=matcher,
                                   distance_thresholds=distance_thresholds,
                                   tb_logdir=log_dir)

    assert tf.math.reduce_all(callback.targets == targets)
    assert tf.math.reduce_all(callback.target_labels == target_labels)
    assert callback.distance == 'l2'
    assert isinstance(callback.evaluator, MemoryEvaluator)
    assert {'recall'} == set([m.name for m in callback.metrics])
    assert callback.k == 11
    assert tf.math.reduce_all(
        callback.distance_thresholds == distance_thresholds)
    assert callback.matcher == 'majority_vote'
    assert isinstance(callback.tb_writer, tf.summary.SummaryWriter)

    assert tf.math.reduce_all(callback.queries_known == q_known)
    assert tf.math.reduce_all(callback.query_labels_known == q_label_known)
    assert tf.math.reduce_all(callback.queries_unknown == q_unknown)
    assert tf.math.reduce_all(callback.query_labels_unknown == q_label_unknown)


def test_split_val_loss_callback(tmp_path):
    queries = tf.constant([[1, 2], [1, 2]])
    query_labels = tf.constant([1, 2])
    targets = tf.constant([[1, 2], [1, 2]])
    target_labels = tf.constant([1, 2])
    known_classes = tf.constant([1])

    log_dir = tmp_path / 'sec/'
    if not log_dir.exists():
        log_dir.mkdir(parents=True)

    callback = SplitValidationLoss(queries=queries,
                                   query_labels=query_labels,
                                   targets=targets,
                                   target_labels=target_labels,
                                   known_classes=known_classes,
                                   tb_logdir=str(log_dir))

    # manually set model ^^
    tf.keras.backend.clear_session()
    inputs = tf.keras.layers.Input(shape=(2,))
    outputs = tf.keras.layers.Dense(2)(inputs)
    model = SimilarityModel(inputs, outputs)
    model.compile('adam', loss='mse', distance='cosine')
    callback.model = model

    # call the only callback method implemented
    callback.on_epoch_end(0, {})



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Classification metrics measure matching classification quality between a
set query examples and a set of indexed examples. """
from .classification_metric import ClassificationMetric  # noqa
from .f1_score import F1Score  # noqa
from .false_positive_rate import FalsePositiveRate  # noqa
from .negative_predictive_value import NegativePredictiveValue  # noqa
from .precision import Precision  # noqa
from .recall import Recall  # noqa
from .binary_accuracy import BinaryAccuracy  # noqa
from .utils import make_classification_metric  # noqa



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"Various utilities functions for improved quality of life."
from typing import Optional, Sequence

import tensorflow as tf

from tensorflow_similarity.types import FloatTensor
from tensorflow_similarity.types import IntTensor
from tensorflow_similarity.types import Lookup


def is_tensor_or_variable(x):
    "check if a variable is tf.Tensor or tf.Variable"
    return tf.is_tensor(x) or isinstance(x, tf.Variable)


def tf_cap_memory():
    "Avoid TF to hog memory before needing it"
    gpus = tf.config.experimental.list_physical_devices('GPU')

    if gpus:
        for gpu in gpus:
            try:
                tf.config.experimental.set_memory_growth(gpu, True)
            except RuntimeError as e:
                # Memory growth must be set before GPUs have been initialized
                print(e)


def unpack_lookup_labels(lookups: Sequence[Sequence[Lookup]]) -> IntTensor:
    # using list comprehension as it is faster
    all_values = [[n.label for n in lu] for lu in lookups]
    result: IntTensor = tf.cast(tf.constant(all_values), dtype='int32')
    return result


def unpack_lookup_distances(
        lookups: Sequence[Sequence[Lookup]],
        distance_rounding: Optional[int] = None) -> FloatTensor:
    # using list comprehension as it is faster
    all_values = [[n.distance for n in lu] for lu in lookups]
    dists: FloatTensor = tf.cast(tf.constant(all_values), dtype='float32')

    if distance_rounding is not None:
        multiplier = tf.constant([10.0**distance_rounding])
        dists = tf.round(dists * multiplier) / multiplier

    return dists



import pytest

from tensorflow_similarity.samplers import TFDatasetMultiShotMemorySampler


def test_basic():
    dataset_name = "mnist"
    sampler = TFDatasetMultiShotMemorySampler(
        dataset_name=dataset_name, classes_per_batch=10
    )
    batch = sampler.generate_batch(42)
    assert batch[0].shape == (20, 28, 28, 1)


def test_wrong_key():
    dataset_name = "mnist"

    # X
    with pytest.raises(ValueError):
        TFDatasetMultiShotMemorySampler(
            dataset_name=dataset_name, classes_per_batch=4, x_key="error"
        )
    # Y
    with pytest.raises(ValueError):
        TFDatasetMultiShotMemorySampler(
            dataset_name=dataset_name, classes_per_batch=4, y_key="error"
        )



import re

import pytest
import tensorflow as tf

from tensorflow_similarity.retrieval_metrics import RetrievalMetric
from tensorflow_similarity.types import FloatTensor, IntTensor, BoolTensor


class ConcreteRetrievalMetric(RetrievalMetric):
    def compute(
        self,
        *,  # keyword only arguments see PEP-570
        query_labels: IntTensor,
        lookup_labels: IntTensor,
        lookup_distances: FloatTensor,
        match_mask: BoolTensor,
    ) -> FloatTensor:
        return tf.constant(1.0)


def test_concrete_instance():
    rm = ConcreteRetrievalMetric(
        name="foo",
        canonical_name="bar",
        k=6,
        average="macro",
    )

    assert rm.name == "foo@6"
    # Check the name once we have updated the threshold
    rm.distance_threshold = 0.1
    assert rm.name == "foo@6 : distance_threshold@0.1"
    assert repr(rm) == "bar : foo@6 : distance_threshold@0.1"
    assert rm.canonical_name == "bar"
    assert rm.k == 6
    assert rm.distance_threshold == 0.1
    assert rm.average == "macro"

    expected_config = {
        "name": "foo@6 : distance_threshold@0.1",
        "canonical_name": "bar",
        "k": 6,
        "distance_threshold": 0.1,
    }
    assert rm.get_config() == expected_config


def test_k_greater_than_num_lookups():
    query_labels = tf.constant([1, 1])
    match_mask = tf.constant(
        [
            [True, True, False, False],
            [True, False, False, True],
        ],
        dtype=bool,
    )
    rm = ConcreteRetrievalMetric(
        name="foo",
        canonical_name="bar",
        k=5,
        average="macro",
    )

    msg = (
        "The number of neighbors must be >= K. Number of neighbors is 4 but "
        "K is 5."
    )

    with pytest.raises(ValueError, match=re.escape(msg)):
        _ = rm._check_shape(query_labels=query_labels, match_mask=match_mask)


def test_query_and_match_mask_different_dims():
    query_labels = tf.constant([1, 2, 3, 4])
    match_mask = tf.constant(
        [
            [True, True, False, False],
            [True, False, False, True],
        ],
        dtype=bool,
    )
    rm = ConcreteRetrievalMetric(
        name="foo",
        canonical_name="bar",
        k=4,
        average="macro",
    )

    msg = (
        "The number of lookup sets must equal the number of query labels. "
        "Number of lookup sets is 2 but the number of query labels is 4."
    )

    with pytest.raises(ValueError, match=re.escape(msg)):
        _ = rm._check_shape(query_labels=query_labels, match_mask=match_mask)



import numpy as np

from tensorflow_similarity.stores import MemoryStore


def build_store(records):
    kv_store = MemoryStore()
    idxs = []
    for r in records:
        idx = kv_store.add(r[0], r[1], r[2])
        idxs.append(idx)
    return kv_store, idxs


def test_in_memory_store_and_retrieve():
    records = [
        [[0.1, 0.2], 1, [0, 0, 0]],
        [[0.2, 0.3], 2, [0, 0, 0]]
    ]

    kv_store, idxs = build_store(records)

    # check index numbering
    for gt, idx in enumerate(idxs):
        assert isinstance(idx, int)
        assert gt == idx

    # check reference counting
    assert kv_store.size() == 2

    # get back three elements
    for idx in idxs:
        emb, lbl, dt = kv_store.get(idx)
        assert emb == records[idx][0]
        assert lbl == records[idx][1]
        assert dt == records[idx][2]


def test_batch_add():
    embs = np.array([[0.1, 0.2], [0.2, 0.3]])
    lbls = np.array([1, 2])
    data = np.array([[0, 0, 0], [1, 1, 1]])

    kv_store = MemoryStore()
    idxs = kv_store.batch_add(embs, lbls, data)
    for idx in idxs:
        emb, lbl, dt = kv_store.get(idx)
        assert np.array_equal(emb, embs[idx])
        assert np.array_equal(lbl, lbls[idx])
        assert np.array_equal(dt, data[idx])


def test_save_and_reload(tmp_path):
    records = [
        [[0.1, 0.2], 1, [0, 0, 0]],
        [[0.2, 0.3], 2, [0, 0, 0]]
    ]

    kv_store, idxs = build_store(records)
    kv_store.save(str(tmp_path))

    # reload
    reloaded_store = MemoryStore()
    reloaded_store.load(tmp_path)

    assert reloaded_store.size() == 2

    # get back three elements
    for idx in idxs:
        emb, lbl, dt = reloaded_store.get(idx)
        assert np.array_equal(emb, records[idx][0])
        assert np.array_equal(lbl, records[idx][1])
        assert np.array_equal(dt, records[idx][2])



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import tensorflow as tf

from tensorflow_similarity.types import IntTensor
from tensorflow_similarity.types import BoolTensor


def compute_match_mask(query_labels: IntTensor,
                       lookup_labels: IntTensor) -> BoolTensor:
    """Compute a boolean mask (indicator function) marking the TPs in the results.

    Args:
        query_labels: A 1D tensor of the labels associated with the queries.

        lookup_labels: A 2D tensor where the jth row is the labels associated
        with the set of k neighbors for the jth query.

    Returns:
        A 2D boolean tensor indicating which lookups match the label of their
        associated query.
    """
    if tf.rank(query_labels) == 1:
        query_labels = tf.expand_dims(query_labels, axis=-1)

    match_mask: BoolTensor = tf.math.equal(lookup_labels, query_labels)

    return match_mask



from unittest import mock
import pytest

import tensorflow as tf
from tensorflow_similarity.visualization import viz_neigbors_imgs
from tensorflow_similarity.visualization.neighbors_viz import _get_class_label
from tensorflow_similarity.types import Lookup


class TestGetClassLabel():

    @pytest.fixture
    def class_mapping(self):
        return {0: 'foo', 1: 'bar'}

    def test_example_class_is_none(self, class_mapping):
        c_lbl = _get_class_label(None, class_mapping)
        assert c_lbl == 'No Label'

    def test_class_mapping_is_none(self):
        c_lbl = _get_class_label(0, None)
        assert c_lbl == '0'

    def test_get_class_label(self, class_mapping):
        c_lbl = _get_class_label(0, class_mapping)
        assert c_lbl == 'foo'

        c_lbl = _get_class_label(1, class_mapping)
        assert c_lbl == 'bar'

    def test_example_class_not_in_mapping(self, class_mapping):
        c_lbl = _get_class_label(2, class_mapping)
        assert c_lbl == '2'

    def test_class_mapping_must_implement_get(self):
        msg = "'list' object has no attribute 'get'"
        with pytest.raises(AttributeError, match=msg):
            _ = _get_class_label(0, ['foo', 'bar'])


@mock.patch('tensorflow_similarity.visualization.neighbors_viz.plt.show')
@mock.patch('tensorflow_similarity.visualization.neighbors_viz.plt.subplots',
            autospec=True)
def test_viz_neighbors_imgs(mock_subplots, mock_show):
    ax_0 = mock.Mock()
    ax_1 = mock.Mock()
    ax_2 = mock.Mock()
    ax_3 = mock.Mock()
    mock_subplots.return_value = (None, [ax_0, ax_1, ax_2, ax_3])

    query_img = tf.constant([1.0])
    nn = [
        Lookup(
            rank=0,  # Incorrect class but class id in class mapping
            distance=0.2,
            label=2,
            data=tf.constant([2.0])),
        Lookup(
            rank=0,  # Incorrect class and class if not in mapping
            distance=0.3,
            label=3,
            data=tf.constant([3.0])),
        Lookup(
            rank=0,  # Correct class and class in mapping
            distance=0.1,
            label=1,
            data=tf.constant([4.0]))
    ]

    viz_neigbors_imgs(example=query_img,
                      example_class=1,
                      neighbors=nn,
                      class_mapping={
                          1: 'foo',
                          2: 'bar'
                      },
                      fig_size=(10, 10),
                      cmap='Blues')

    mock_subplots.assert_called_with(nrows=1, ncols=4, figsize=(10, 10))

    ax_0.imshow.assert_called_with(query_img, cmap='Blues')
    ax_0.set_xticks.assert_called_with([])
    ax_0.set_yticks.assert_called_with([])
    ax_0.set_title.assert_called_with('foo')

    ax_1.imshow.assert_called_with(tf.constant([2.0]), cmap='Reds')
    ax_1.set_xticks.assert_called_with([])
    ax_1.set_yticks.assert_called_with([])
    ax_1.set_title.assert_called_with('bar - 0.20000')

    ax_2.imshow.assert_called_with(tf.constant([3.0]), cmap='Reds')
    ax_2.set_xticks.assert_called_with([])
    ax_2.set_yticks.assert_called_with([])
    ax_2.set_title.assert_called_with('3 - 0.30000')

    ax_3.imshow.assert_called_with(tf.constant([4.0]), cmap='Blues')
    ax_3.set_xticks.assert_called_with([])
    ax_3.set_yticks.assert_called_with([])
    ax_3.set_title.assert_called_with('foo - 0.10000')



import numpy as np
from tensorflow_similarity.indexer import Indexer
from . import DATA_DIR


def test_calibration():
    # CALIB TEST
    SIZE = 20

    FNAME = str(DATA_DIR / 'mnist_fashion_embeddings.npz')
    data = np.load(FNAME, allow_pickle=True)
    thresholds_targets = {'0.5': 0.5}

    index = Indexer(3)
    index.batch_add(data['embeddings_idx'][:SIZE], labels=data['y_idx'][:SIZE])
    calibration = index.calibrate(data['embeddings_cal'][:SIZE],
                                  data['y_cal'][:SIZE],
                                  thresholds_targets,
                                  verbose=1)
    # assert 'vl' in cutpoints
    assert 'optimal' in calibration.cutpoints
    assert '0.5' in calibration.cutpoints
    assert len(calibration.thresholds['distance']) == len(
        calibration.thresholds['value'])
    assert index.is_calibrated


def test_indexer_basic_flow():

    prediction = np.array([[1, 1, 2]], dtype='float32')
    embs = np.array([[1, 1, 3], [3, 1, 2]], dtype='float32')

    indexer = Indexer(3)

    # index data
    indexer.batch_add(embs, labels=[0, 1], data=['test', 'test2'])

    # lookup
    matches = indexer.single_lookup(prediction)

    assert isinstance(matches, list)
    assert matches[0].distance < 0.016

    assert np.array_equal(matches[0].embedding, embs[0])
    assert matches[0].label == 0
    assert matches[0].data == 'test'


def test_indexer_batch_add():

    prediction = np.array([[1, 1, 2]], dtype='float32')
    embs = np.array([[1, 1, 3], [3, 1, 2]], dtype='float32')

    indexer = Indexer(3)

    # index data
    indexer.batch_add(embs, [0, 1], data=['test', 'test2'])
    assert indexer.size() == 2
    # check results
    matches = indexer.single_lookup(prediction)

    assert isinstance(matches, list)
    assert matches[0].distance < 0.016

    assert np.array_equal(matches[0].embedding, embs[0])
    assert matches[0].label == 0
    assert matches[0].data == 'test'


def test_multiple_add():

    # arrays of preds which contains a single embedding list(list(embedding))
    predictions = np.array([[[1, 1, 3]], [[3, 1, 2]]], dtype='float32')

    indexer = Indexer(3)
    indexer.add(predictions[0])
    assert indexer.size() == 1

    indexer.add(predictions[1])
    assert indexer.size() == 2


def test_multiple_add_mix_data():

    embs = np.array([[1, 1, 3], [3, 1, 2]], dtype='float32')

    indexer = Indexer(3)
    indexer.batch_add(embs)
    assert indexer.size() == 2

    indexer.batch_add(embs, data=embs)
    assert indexer.size() == 4


def test_reload(tmp_path):

    embs = np.array([[1, 1, 3], [3, 1, 2]], dtype='float32')

    indexer = Indexer(3)
    indexer.batch_add(embs, verbose=0)
    assert indexer.size() == 2

    # save
    path = tmp_path / "test_save_and_add/"
    indexer.save(path)

    # reload
    indexer2 = Indexer.load(path)
    assert indexer2.size() == 2

    # add more
    indexer2.batch_add(embs, data=embs)
    assert indexer2.size() == 4


def test_uncompress_reload(tmp_path):
    "Ensure uncompressed index work"

    embs = np.array([[1, 1, 3], [3, 1, 2]], dtype='float32')

    indexer = Indexer(3)
    indexer.batch_add(embs, verbose=0)
    assert indexer.size() == 2

    # save
    path = tmp_path / "test_save_and_add/"
    indexer.save(path, compression=False)

    # reload
    indexer2 = Indexer.load(path)
    assert indexer2.size() == 2


def test_index_reset():

    prediction = np.array([[1, 1, 2]], dtype='float32')
    embs = np.array([[1, 1, 3], [3, 1, 2], [3, 2, 3]], dtype='float32')

    indexer = Indexer(3)

    # index data
    indexer.batch_add(embs, labels=[0, 1, 2])

    # lookup
    matches = indexer.single_lookup(prediction)

    # get stats
    stats = indexer.stats()

    # check results
    assert len(matches) == 3
    assert len(matches[0].embedding) == 3
    assert matches[0].label == 0
    assert list(matches[0].embedding) == list(embs[0])

    # reset
    indexer.reset()
    stats = indexer.stats()
    assert stats['num_lookups'] == 0
    assert stats['num_items'] == 0

    # do-over
    indexer.add([embs[0]], label=42)
    indexer.add([embs[1]], label=43)

    matches = indexer.single_lookup(prediction)
    stats = indexer.stats()

    assert len(matches) == 2
    assert len(matches[0].embedding) == 3
    assert matches[0].label == 42
    assert matches[1].label == 43
    assert list(matches[0].embedding) == list(embs[0])
    assert list(matches[1].embedding) == list(embs[1])
    assert stats['num_lookups'] == 1


def test_indexer_batch_ops():
    NUM_ELTS = 100
    NUM_DIMS = 10
    K = 3
    data = np.random.randn(NUM_ELTS, NUM_DIMS).astype(np.float32)
    indexer = Indexer(NUM_DIMS)
    indexer.batch_add(data)
    results = indexer.batch_lookup(data, k=K)
    indexer.stats()
    assert len(results) == 100
    assert len(results[0]) == K


def test_single_vs_batch_ops():
    "ensure batch and single ops are consistent"
    NUM_ELTS = 100
    NUM_DIMS = 10
    K = 3
    data = np.random.randn(NUM_ELTS, NUM_DIMS).astype(np.float32)
    indexer = Indexer(NUM_DIMS)
    indexer.batch_add(data)
    batch_results = indexer.batch_lookup(data, k=K)

    single_results = []
    for d in data:
        single_results.append(indexer.single_lookup([d], k=K))

    for idx in range(len(single_results)):
        assert single_results[idx][0].label == batch_results[idx][0].label



# Copyright 2021 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
from typing import Any, Callable, Optional, Tuple

from tensorflow.keras.utils import Sequence
from tensorflow_similarity.types import FloatTensor, IntTensor

# An Augmenter is a Map TensorLike -> TensorLike. The function must
# implement the following signature:
#
# Args:
#   x: Feature Tensors
#   y: Label Tensors
#   examples_per_class: The number of examples per class.
#   is_warmup: If True, the sampler is still in warmup phase.
# Returns:
#   A Tuple containing the transformed x and y tensors.
Augmenter = Callable[[FloatTensor, IntTensor, int, bool], Tuple[FloatTensor,
                                                                IntTensor]]

# Not currently used. Might be useful to allows gradual call of augmenter
Scheduler = Callable[[Any], Any]


class Sampler(Sequence, metaclass=abc.ABCMeta):

    def __init__(
        self,
        classes_per_batch: int,
        examples_per_class_per_batch: int = 2,
        num_augmentations_per_example: int = 0,
        steps_per_epoch: int = 1000,
        augmenter: Optional[Augmenter] = None,
        # scheduler: Optional[Scheduler] = None,
        warmup: int = 0,
    ) -> None:
        """Create a dataset sampler that ensure that each batch contains at
        least `examples_per_class_per_batch` examples of `classes_per_batch`
        classes. Sampling is needed as contrastive loss requires at least two
        examples of the same class present in a batch. The batch_size used
        during training will be equal to: `classes_per_batch` *
        `examples_per_class_per_batch` unless an `augmenter` that alters the
        number of examples returned is used. Then the batch_size is a function
        of how many augmented examples are returned by the `augmenter`.


        Args:
            classes_per_batch: Numbers of classes to include in a single batch

            examples_per_class_per_batch: how many examples of each class to use
            per batch. Defaults to 2.

            num_augmentations_per_example: how many augmented versions of an
            example will be produced by the augmenter. Defaults to 0.

            steps_per_epoch: How many steps/batches per epoch. Defaults to
            1000.

            augmenter: A function that takes a batch in and returns a batch out.
            Can alter the number of examples returned, which in turn can change
            the batch_size used. Defaults to None.

            warmup: Keep track of warmup epochs and let the augmenter know
            when the warmup is over by passing along with each batch a boolean
            `is_warmup`. See `self._get_examples()` Defaults to 0.
        """

        self.epoch = 0  # track epoch count
        self.classes_per_batch = classes_per_batch
        self.examples_per_class_per_batch = examples_per_class_per_batch
        self.num_augmentations_per_example = num_augmentations_per_example
        self.batch_size = classes_per_batch * examples_per_class_per_batch
        self.aug_size = (self.batch_size *
                         (1 + num_augmentations_per_example)) - self.batch_size
        self.steps_per_epoch = steps_per_epoch
        self.augmenter = augmenter
        self.warmup = warmup
        self.is_warmup = True if warmup else False

        # Tell the users what to expect as they might be unsure what the batch
        # size will be
        print(f"\nThe initial batch size is {self.batch_size + self.aug_size} "
              f"({self.classes_per_batch} classes * "
              f"{self.examples_per_class_per_batch} examples per class) with "
              f"{self.num_augmentations_per_example} augmenters")

    @abc.abstractmethod
    def _get_examples(self, batch_id: int, num_classes: int,
                      examples_per_class: int) -> Tuple[FloatTensor, IntTensor]:
        """Get the set of examples that would be used to create a single batch.

        Notes:
         - before passing the batch data to TF, the sampler will call the
           augmenter function (if any) on the returned example.

         - A batch_size = num_classes * examples_per_class

         - This function must be defined in the subclass.

        Args:
            batch_id: id of the batch in the epoch.

            num_classes: How many class should be present in the examples.

            examples_per_class: How many example per class should be returned.

        Returns:
            x, y: batch of examples size `num_classes` * `examples_per_class`
        """

    # [Shared mechanics]
    def __len__(self) -> int:
        "Return the number of batch per epoch"
        return self.steps_per_epoch

    def on_epoch_end(self) -> None:
        "Keep track of warmup epochs"

        # if self.scheduler:
        #     # fixme scheduler idea
        #     self.scheduler(self.epoch)

        self.epoch += 1
        if self.is_warmup and (self.epoch >= self.warmup):
            print("Warmup complete")
            self.is_warmup = False

    def __getitem__(self, batch_id: int) -> Tuple[FloatTensor, IntTensor]:
        return self.generate_batch(batch_id)

    def generate_batch(self, batch_id: int) -> Tuple[FloatTensor, IntTensor]:
        """Generate a batch of data.


        Args:
            batch_id ([type]): [description]

        Returns:
            x, y: batch
        """

        x, y = self._get_examples(batch_id, self.classes_per_batch,
                                  self.examples_per_class_per_batch)

        # strip examples if needed. This might happen due to rounding
        if len(x) > self.batch_size:
            x = x[:self.batch_size]
            y = y[:self.batch_size]

        if self.augmenter:
            x, y = self.augmenter(x, y, self.num_augmentations_per_example,
                                  self.is_warmup)
        return x, y
